from .deficit_model import DeficitModel
from .selfsimilarity import SelfSimilarityDeficit
from .noj import NOJDeficit, NOJLocalDeficit
from .gaussian import BastankhahGaussianDeficit, IEA37SimpleBastankhahGaussianDeficit, \
    NiayifarGaussianDeficit, ZongGaussianDeficit
from .fuga import FugaDeficit
from .gcl import GCLDeficit
